@@ -2,12 +2,13 @@
 #include <sys/select.h>
 #include <sys/time.h>
 
+#include <ctype.h>
 #include <locale.h>
 #include <signal.h>
 #include <stdarg.h>
 #include <stdio.h>
-#include <string.h>
 #include <stdlib.h>
+#include <string.h>
 #include <time.h>
 #include <unistd.h>
 
@@ -20,6 +21,7 @@
 #include <X11/Xproto.h>
 #include <X11/extensions/XTest.h>
 #include <X11/Xft/Xft.h>
+#include <X11/Xresource.h>
 #ifdef XINERAMA
 #include <X11/extensions/Xinerama.h>
 #endif
@@ -29,21 +31,28 @@
 
 /* macros */
 #define LENGTH(x)         (sizeof x / sizeof x[0])
-#define TEXTW(X)          (drw_fontset_getwidth(drw, (X)))
 #define STRINGTOKEYSYM(X) (XStringToKeySym(X))
+#define TEXTW(X)          (drw_fontset_getwidth(drw, (X)))
 
 /* enums */
-enum { SchemeNorm, SchemePress, SchemeHighlight, SchemeLast };
+enum {
+	SchemeNorm, SchemeNormABC, SchemeNormABCShift, SchemeNormShift, SchemePress,
+	SchemePressShift, SchemeHighlight, SchemeHighlightShift, SchemeOverlay,
+	SchemeOverlayShift, SchemeWindow, SchemeLast
+};
 enum { NetWMWindowType, NetLast };
 
 /* typedefs */
 typedef struct {
 	char *label;
+	char *label2;
 	KeySym keysym;
-	unsigned int width;
+	double width;
+	KeySym modifier;
 	int x, y, w, h;
 	Bool pressed;
 	Bool highlighted;
+	Bool isoverlay;
 } Key;
 
 typedef struct {
@@ -61,11 +70,11 @@ static void configurenotify(XEvent *e);
 static void countrows();
 static int countkeys(Key *layer);
 static void drawkeyboard(void);
-static void drawkey(Key *k);
+static void drawkey(Key *k, Bool map);
 static void expose(XEvent *e);
 static Key *findkey(int x, int y);
 static void leavenotify(XEvent *e);
-static void press(Key *k, KeySym mod);
+static void press(Key *k, KeySym buttonmod);
 static double get_press_duration();
 static void run(void);
 static void setup(void);
@@ -76,8 +85,9 @@ static void hideoverlay();
 static void cyclelayer();
 static void setlayer();
 static void togglelayer();
-static void unpress(Key *k, KeySym mod);
+static void unpress(Key *k, KeySym buttonmod);
 static void updatekeys();
+static void printkey(Key *k, KeySym mod);
 
 /* variables */
 static int screen;
@@ -95,20 +105,25 @@ static Drw *drw;
 static Window root, win;
 static Clr* scheme[SchemeLast];
 static Bool running = True, isdock = False;
-static KeySym pressedmod = 0;
 static struct timeval pressbegin;
 static int currentlayer = 0;
 static int enableoverlays = 1;
 static int currentoverlay = -1; /* -1 = no overlay */
+static int pressonrelease = 1;
 static KeySym overlaykeysym = 0; /* keysym for which the overlay is presented */
 static int releaseprotect = 0; /* set to 1 after overlay is shown, protecting against immediate release */
 static int tmp_keycode = 1;
 static int rows = 0, ww = 0, wh = 0, wx = 0, wy = 0;
+static int simulateoutput = 1; /* simulate key presses for X */
+static int printoutput = 0; /* print key pressed to stdout */
 static char *name = "svkbd";
 static int debug = 0;
 static int numlayers = 0;
 static int numkeys = 0;
 
+static char *colors[11][2]; /* 11 schemes, 2 colors each */
+static char *fonts[] = { 0 };
+
 static KeySym ispressingkeysym;
 
 Bool ispressing = False;
@@ -121,14 +136,16 @@ Bool sigtermd = False;
 #endif
 #include LAYOUT
 
-static Key keys[KEYS] = { NULL };
-static Key* layers[LAYERS];
+static Key keys[KEYS];
+static Key *layers[LAYERS];
 
 void
 motionnotify(XEvent *e)
 {
 	XPointerMovedEvent *ev = &e->xmotion;
 	int i;
+	int lostfocus = -1;
+	int gainedfocus = -1;
 
 	for (i = 0; i < numkeys; i++) {
 		if (keys[i].keysym && ev->x > keys[i].x
@@ -137,25 +154,36 @@ motionnotify(XEvent *e)
 				&& ev->y < keys[i].y + keys[i].h) {
 			if (keys[i].highlighted != True) {
 				if (ispressing) {
-					keys[i].pressed = True;
+					gainedfocus = i;
 				} else {
 					keys[i].highlighted = True;
 				}
-				drawkey(&keys[i]);
+				drawkey(&keys[i], True);
 			}
 			continue;
-		}
-
-		if (!IsModifierKey(keys[i].keysym) && keys[i].pressed == True) {
-			unpress(&keys[i], 0);
-
-			drawkey(&keys[i]);
-		}
-		if (keys[i].highlighted == True) {
+		} else if (keys[i].highlighted == True) {
 			keys[i].highlighted = False;
-			drawkey(&keys[i]);
+			drawkey(&keys[i], True);
 		}
 	}
+
+	for (i = 0; i < numkeys; i++) {
+		if (!IsModifierKey(keys[i].keysym) && keys[i].pressed == True &&
+		    lostfocus != gainedfocus) {
+			printdbg("Pressed key lost focus: %ld\n", keys[i].keysym);
+			lostfocus = i;
+			ispressingkeysym = 0;
+			keys[i].pressed = 0;
+			drawkey(&keys[i], True);
+		}
+	}
+
+	if ((lostfocus != -1) && (gainedfocus != -1) && (lostfocus != gainedfocus)) {
+		printdbg("Clicking new key that gained focus\n");
+		press(&keys[gainedfocus], 0);
+		keys[gainedfocus].pressed = True;
+		keys[gainedfocus].highlighted = True;
+	}
 }
 
 void
@@ -168,14 +196,23 @@ buttonpress(XEvent *e)
 
 	ispressing = True;
 
+	if (!(k = findkey(ev->x, ev->y)))
+		return;
+
 	for (i = 0; i < LENGTH(buttonmods); i++) {
 		if (ev->button == buttonmods[i].button) {
 			mod = buttonmods[i].mod;
 			break;
 		}
 	}
-	if ((k = findkey(ev->x, ev->y)))
-		press(k, mod);
+
+	if (k->modifier) {
+		if (mod == k->modifier)
+			mod = 0;
+		else
+			mod = k->modifier;
+	}
+	press(k, mod);
 }
 
 void
@@ -197,8 +234,12 @@ buttonrelease(XEvent *e)
 
 	if (ev->x < 0 || ev->y < 0) {
 		unpress(NULL, mod);
-	} else {
-		if ((k = findkey(ev->x, ev->y)))
+	} else if ((k = findkey(ev->x, ev->y))) {
+		if (k->modifier == mod)
+			unpress(k, 0);
+		else if (k->modifier)
+			unpress(k, k->modifier);
+		else
 			unpress(k, mod);
 	}
 }
@@ -213,7 +254,6 @@ cleanup(void)
 	drw_sync(drw);
 	drw_free(drw);
 	XSync(dpy, False);
-	drw_free(drw);
 	XDestroyWindow(dpy, win);
 	XSync(dpy, False);
 	XSetInputFocus(dpy, PointerRoot, RevertToPointerRoot, CurrentTime);
@@ -238,8 +278,13 @@ countrows(void)
 	int i;
 
 	for (i = 0, rows = 1; i < numkeys; i++) {
-		if (keys[i].keysym == 0)
+		if (keys[i].keysym == 0) {
 			rows++;
+			if ((i > 0) && (keys[i-1].keysym == 0)) {
+				rows--;
+				break;
+			}
+		}
 	}
 }
 
@@ -264,26 +309,38 @@ drawkeyboard(void)
 {
 	int i;
 
+	drw_setscheme(drw, scheme[SchemeWindow]);
+	drw_rect(drw, 0, 0, ww, wh, 1, 1);
 	for (i = 0; i < numkeys; i++) {
 		if (keys[i].keysym != 0)
-			drawkey(&keys[i]);
+			drawkey(&keys[i], False);
 	}
+	drw_map(drw, win, 0, 0, ww, wh);
 }
 
 void
-drawkey(Key *k)
+drawkey(Key *k, Bool map)
 {
 	int x, y, w, h;
 	const char *l;
 
+	int use_scheme = SchemeNorm;
+
 	if (k->pressed)
-		drw_setscheme(drw, scheme[SchemePress]);
+		use_scheme = SchemePress;
 	else if (k->highlighted)
-		drw_setscheme(drw, scheme[SchemeHighlight]);
+		use_scheme = SchemeHighlight;
+	else if (k->isoverlay)
+		use_scheme = SchemeOverlay;
+	else if ((k->keysym == XK_Return) ||
+			((k->keysym >= XK_a) && (k->keysym <= XK_z)) ||
+			((k->keysym >= XK_Cyrillic_io) && (k->keysym <= XK_Cyrillic_hardsign)))
+		use_scheme = SchemeNormABC;
 	else
-		drw_setscheme(drw, scheme[SchemeNorm]);
+		use_scheme = SchemeNorm;
+
+	drw_setscheme(drw, scheme[use_scheme]);
 	drw_rect(drw, k->x, k->y, k->w, k->h, 1, 1);
-	drw_rect(drw, k->x, k->y, k->w, k->h, 0, 0);
 
 	if (k->keysym == XK_KP_Insert) {
 		if (enableoverlays) {
@@ -301,7 +358,26 @@ drawkey(Key *k)
 	w = TEXTW(l);
 	x = k->x + (k->w / 2) - (w / 2);
 	drw_text(drw, x, y, w, h, 0, l, 0);
-	drw_map(drw, win, k->x, k->y, k->w, k->h);
+	if (k->label2) {
+		if (use_scheme == SchemeNorm)
+			use_scheme = SchemeNormShift;
+		else if (use_scheme == SchemeNormABC)
+			use_scheme = SchemeNormABCShift;
+		else if (use_scheme == SchemePress)
+			use_scheme = SchemePressShift;
+		else if (use_scheme == SchemeHighlight)
+			use_scheme = SchemeHighlightShift;
+		else if (use_scheme == SchemeOverlay)
+			use_scheme = SchemeOverlayShift;
+		drw_setscheme(drw, scheme[use_scheme]);
+		x += w;
+		y -= 15;
+		l = k->label2;
+		w = TEXTW(l);
+		drw_text(drw, x, y, w, h, 0, l, 0);
+	}
+	if (map)
+		drw_map(drw, win, k->x, k->y, k->w, k->h);
 }
 
 void
@@ -335,9 +411,9 @@ hasoverlay(KeySym keysym)
 	begin = 0;
 	for (i = 0; i < OVERLAYS; i++) {
 		if (overlay[i].keysym == XK_Cancel) {
-			begin = i+1;
+			begin = i + 1;
 		} else if (overlay[i].keysym == keysym) {
-			return begin+1;
+			return begin + 1;
 		}
 	}
 	return -1;
@@ -348,6 +424,7 @@ leavenotify(XEvent *e)
 {
 	if (currentoverlay != -1)
 		hideoverlay();
+	ispressingkeysym = 0;
 	unpress(NULL, 0);
 }
 
@@ -360,14 +437,14 @@ record_press_begin(KeySym ks)
 }
 
 void
-press(Key *k, KeySym mod)
+press(Key *k, KeySym buttonmod)
 {
 	int i;
 	int overlayidx = -1;
 
 	k->pressed = !k->pressed;
 
-	if (debug) printdbg("Begin press: %ld\n", k->keysym);
+	printdbg("Begin click: %ld\n", k->keysym);
 	pressbegin.tv_sec = 0;
 	pressbegin.tv_usec = 0;
 	ispressingkeysym = 0;
@@ -375,23 +452,21 @@ press(Key *k, KeySym mod)
 	if (!IsModifierKey(k->keysym)) {
 		if (enableoverlays && currentoverlay == -1)
 			overlayidx = hasoverlay(k->keysym);
-		if (enableoverlays && overlayidx != -1) {
-			if (!pressbegin.tv_sec && !pressbegin.tv_usec) {
-				/*record the begin of the press, don't simulate the actual keypress yet */
-				record_press_begin(k->keysym);
-			}
+		if ((pressonrelease) || (enableoverlays && overlayidx != -1)) {
+			/*record the begin of the press, don't simulate the actual keypress yet */
+			record_press_begin(k->keysym);
 		} else {
-			if (debug) printdbg("Simulating press: %ld\n", k->keysym);
+			printdbg("Simulating press: %ld (mod %ld)\n", k->keysym, buttonmod);
 			for (i = 0; i < numkeys; i++) {
 				if (keys[i].pressed && IsModifierKey(keys[i].keysym)) {
 					simulate_keypress(keys[i].keysym);
 				}
 			}
-			pressedmod = mod;
-			if (pressedmod) {
-				simulate_keypress(mod);
-			}
+			if (buttonmod)
+				simulate_keypress(buttonmod);
 			simulate_keypress(k->keysym);
+			if (printoutput)
+				printkey(k, buttonmod);
 
 			for (i = 0; i < numkeys; i++) {
 				if (keys[i].pressed && IsModifierKey(keys[i].keysym)) {
@@ -400,7 +475,7 @@ press(Key *k, KeySym mod)
 			}
 		}
 	}
-	drawkey(k);
+	drawkey(k, True);
 }
 
 int
@@ -412,10 +487,58 @@ tmp_remap(KeySym keysym)
 	return tmp_keycode;
 }
 
+void
+printkey(Key *k, KeySym mod)
+{
+	int i, shift;
+
+	shift = (mod == XK_Shift_L) || (mod == XK_Shift_R) || (mod == XK_Shift_Lock);
+	if (!shift) {
+		for (i = 0; i < numkeys; i++) {
+			if ((keys[i].pressed) && ((keys[i].keysym == XK_Shift_L) ||
+			    (keys[i].keysym == XK_Shift_R) || (keys[i].keysym == XK_Shift_Lock))) {
+				shift = True;
+				break;
+			}
+		}
+	}
+	printdbg("Printing key %ld (shift=%d)\n", k->keysym, shift);
+	if (k->keysym == XK_Cancel)
+		return;
+
+	KeySym *keysym = &(k->keysym);
+	XIM xim = XOpenIM(dpy, 0, 0, 0);
+	XIC xic = XCreateIC(xim, XNInputStyle, XIMPreeditNothing | XIMStatusNothing, NULL);
+
+	XKeyPressedEvent event;
+	event.type = KeyPress;
+	event.display = dpy;
+	event.state = shift ? ShiftMask : 0;
+	event.keycode = XKeysymToKeycode(dpy, *keysym);
+	if (event.keycode == 0)
+		event.keycode = tmp_remap(*keysym);
+
+	char buffer[32];
+	KeySym ignore;
+	Status return_status;
+	int l = Xutf8LookupString(xic, &event, buffer, sizeof(buffer), &ignore, &return_status);
+	buffer[l] = '\0';
+	printdbg("Print buffer: [%s] (length=%d)\n", &buffer, l);
+	printf("%s", buffer);
+
+	XDestroyIC(xic);
+	XCloseIM(xim);
+}
+
 void
 simulate_keypress(KeySym keysym)
 {
-	KeyCode code = XKeysymToKeycode(dpy, keysym);
+	KeyCode code;
+
+	if (!simulateoutput)
+		return;
+
+	code = XKeysymToKeycode(dpy, keysym);
 	if (code == 0)
 		code = tmp_remap(keysym);
 	XTestFakeKeyEvent(dpy, code, True, 0);
@@ -424,7 +547,12 @@ simulate_keypress(KeySym keysym)
 void
 simulate_keyrelease(KeySym keysym)
 {
-	KeyCode code = XKeysymToKeycode(dpy, keysym);
+	KeyCode code;
+
+	if (!simulateoutput)
+		return;
+
+	code = XKeysymToKeycode(dpy, keysym);
 	if (code == 0)
 		code = tmp_remap(keysym);
 	XTestFakeKeyEvent(dpy, code, False, 0);
@@ -438,16 +566,17 @@ get_press_duration(void)
 	gettimeofday(&now, NULL);
 
 	return (double) ((now.tv_sec * 1000000L + now.tv_usec) -
-	       (pressbegin.tv_sec * 1000000L + pressbegin.tv_usec)) /
-	       (double) 1000000L;
+		   (pressbegin.tv_sec * 1000000L + pressbegin.tv_usec)) /
+		   (double) 1000000L;
 }
 
 void
-unpress(Key *k, KeySym mod)
+unpress(Key *k, KeySym buttonmod)
 {
 	int i;
+	Bool neutralizebuttonmod = False;
 
-	if (k != NULL) {
+	if (k) {
 		switch(k->keysym) {
 		case XK_Cancel:
 			cyclelayer();
@@ -466,62 +595,57 @@ unpress(Key *k, KeySym mod)
 		}
 	}
 
-	if ((pressbegin.tv_sec || pressbegin.tv_usec) && enableoverlays && k && k->keysym == ispressingkeysym) {
-		if (currentoverlay == -1) {
-			if (get_press_duration() < overlay_delay) {
-				if (debug) printdbg("Delayed simulation of press after release: %ld\n", k->keysym);
-				/* simulate the press event, as we postponed it earlier in press() */
-				for (i = 0; i < numkeys; i++) {
-					if (keys[i].pressed && IsModifierKey(keys[i].keysym)) {
-						simulate_keypress(keys[i].keysym);
-					}
-				}
-				pressedmod = mod;
-				if (pressedmod) {
-					simulate_keypress(mod);
-				}
-				simulate_keypress(k->keysym);
-				pressbegin.tv_sec = 0;
-				pressbegin.tv_usec = 0;
-			} else {
-				return;
+	if ((pressbegin.tv_sec || pressbegin.tv_usec) && (enableoverlays || pressonrelease) && k && k->keysym == ispressingkeysym) {
+		printdbg("Delayed simulation of press after release: %ld\n", k->keysym);
+		/* simulate the press event, as we postponed it earlier in press() */
+		for (i = 0; i < numkeys; i++) {
+			if (keys[i].pressed && IsModifierKey(keys[i].keysym)) {
+				if (keys[i].keysym == buttonmod)
+					neutralizebuttonmod = True;
+				else
+					simulate_keypress(keys[i].keysym);
 			}
 		}
-	}
-
-	if (debug) {
-		if (k) {
-			printdbg("Simulation of release: %ld\n", k->keysym);
-		} else {
-			printdbg("Simulation of release (all keys)\n");
+		if (buttonmod && !neutralizebuttonmod) {
+			simulate_keypress(buttonmod);
 		}
+		simulate_keypress(k->keysym);
+		pressbegin.tv_sec = 0;
+		pressbegin.tv_usec = 0;
 	}
 
+	if (k)
+		printdbg("Simulation of release: %ld\n", k->keysym);
+	else
+		printdbg("Simulation of release (all keys)\n");
 
 	for (i = 0; i < numkeys; i++) {
 		if (keys[i].pressed && !IsModifierKey(keys[i].keysym)) {
 			simulate_keyrelease(keys[i].keysym);
+			if (printoutput)
+				printkey(&keys[i], buttonmod);
 			keys[i].pressed = 0;
-			drawkey(&keys[i]);
-			break;
+			drawkey(&keys[i], True);
 		}
 	}
-	if (i != numkeys) {
-		if (pressedmod) {
-			simulate_keyrelease(mod);
-		}
-		pressedmod = 0;
 
+	if (buttonmod && !neutralizebuttonmod) {
+		simulate_keyrelease(buttonmod);
+	}
+
+	if (k == NULL || !IsModifierKey(k->keysym)) {
 		for (i = 0; i < numkeys; i++) {
-			if (keys[i].pressed) {
-				simulate_keyrelease(keys[i].keysym);
+			if (keys[i].pressed && IsModifierKey(keys[i].keysym)) {
+				if (!(keys[i].keysym == buttonmod && neutralizebuttonmod))
+					simulate_keyrelease(keys[i].keysym);
 				keys[i].pressed = 0;
-				drawkey(&keys[i]);
+				drawkey(&keys[i], True);
 			}
 		}
 	}
 
-	if (enableoverlays && currentoverlay != -1) {
+	if (enableoverlays && currentoverlay != -1 &&
+	    (k == NULL || !IsModifierKey(k->keysym))) {
 		if (releaseprotect) {
 			releaseprotect = 0;
 		} else {
@@ -537,12 +661,13 @@ run(void)
 	int xfd;
 	fd_set fds;
 	struct timeval tv;
-	double duration = 0.0;
+	double duration;
+	int overlayidx;
 	int i, r;
 
 	xfd = ConnectionNumber(dpy);
-	tv.tv_sec = 1;
-	tv.tv_usec = 0;
+	tv.tv_sec = 0;
+	tv.tv_usec = scan_rate;
 
 	XFlush(dpy);
 
@@ -562,25 +687,33 @@ run(void)
 			/* time-out expired without anything interesting happening, check for long-presses */
 			if (ispressing && ispressingkeysym) {
 				duration = get_press_duration();
-				if (debug == 2) printdbg("%f\n", duration);
-				if (get_press_duration() >= overlay_delay) {
-					if (debug) printdbg("press duration %f\n", duration);
-					showoverlay(hasoverlay(ispressingkeysym));
+				if (debug >= 2)
+					printdbg("%f\n", duration);
+				overlayidx = hasoverlay(ispressingkeysym);
+				duration = get_press_duration();
+				if (overlayidx != -1 && duration >= overlay_delay) {
+					printdbg("press duration %f, activating overlay\n", duration);
+					showoverlay(overlayidx);
 					pressbegin.tv_sec = 0;
 					pressbegin.tv_usec = 0;
 					ispressingkeysym = 0;
+				} else if ((overlayidx == -1) && (duration >= repeat_delay)) {
+					printdbg("press duration %f, activating repeat\n", duration);
+					simulate_keyrelease(ispressingkeysym);
+					simulate_keypress(ispressingkeysym);
+					XSync(dpy, False);
 				}
 			}
 		}
 		if (r == -1 || sigtermd) {
 			/* an error occurred or we received a signal */
 			/* E.g. Generally in scripts we want to call SIGTERM on svkbd in which case
-			        if the user is holding for example the enter key (to execute
-			        the kill or script that does the kill), that causes an issue
-			        since then X doesn't know the keyup is never coming.. (since
-			        process will be dead before finger lifts - in that case we
-			        just trigger out fake up presses for all keys */
-			if (debug) printdbg("signal received, releasing all keys");
+					if the user is holding for example the enter key (to execute
+					the kill or script that does the kill), that causes an issue
+					since then X doesn't know the keyup is never coming.. (since
+					process will be dead before finger lifts - in that case we
+					just trigger out fake up presses for all keys */
+			printdbg("signal received, releasing all keys");
 			for (i = 0; i < numkeys; i++) {
 				XTestFakeKeyEvent(dpy, XKeysymToKeycode(dpy, keys[i].keysym), False, 0);
 			}
@@ -589,6 +722,80 @@ run(void)
 	}
 }
 
+void
+readxresources(void)
+{
+	XrmDatabase xdb;
+	XrmValue xval;
+	char *type, *xrm;
+
+	XrmInitialize();
+
+	if ((xrm = XResourceManagerString(drw->dpy))) {
+		xdb = XrmGetStringDatabase(xrm);
+
+		if (XrmGetResource(xdb, "svkbd.font", "*", &type, &xval) && !fonts[0])
+			fonts[0] = estrdup(xval.addr);
+
+		if (XrmGetResource(xdb, "svkbd.background", "*", &type, &xval) && !colors[SchemeNorm][ColBg])
+			colors[SchemeNorm][ColBg] = estrdup(xval.addr);
+		if (XrmGetResource(xdb, "svkbd.foreground", "*", &type, &xval) && !colors[SchemeNorm][ColFg])
+			colors[SchemeNorm][ColFg] = estrdup(xval.addr);
+
+		if (XrmGetResource(xdb, "svkbd.shiftforeground", "*", &type, &xval) && !colors[SchemeNormShift][ColFg])
+			colors[SchemeNormShift][ColFg] = estrdup(xval.addr);
+		if (XrmGetResource(xdb, "svkbd.shiftbackground", "*", &type, &xval) && !colors[SchemeNormShift][ColBg])
+			colors[SchemeNormShift][ColBg] = estrdup(xval.addr);
+
+		if (XrmGetResource(xdb, "svkbd.ABCforeground", "*", &type, &xval) && !colors[SchemeNormABC][ColFg])
+			colors[SchemeNormABC][ColFg] = estrdup(xval.addr);
+		if (XrmGetResource(xdb, "svkbd.ABCbackground", "*", &type, &xval) && !colors[SchemeNormABC][ColBg])
+			colors[SchemeNormABC][ColBg] = estrdup(xval.addr);
+
+		if (XrmGetResource(xdb, "svkbd.ABCshiftforeground", "*", &type, &xval) && !colors[SchemeNormShift][ColFg])
+			colors[SchemeNormShift][ColFg] = estrdup(xval.addr);
+		if (XrmGetResource(xdb, "svkbd.ABCshiftbackground", "*", &type, &xval) && !colors[SchemeNormShift][ColBg])
+			colors[SchemeNormShift][ColBg] = estrdup(xval.addr);
+
+		if (XrmGetResource(xdb, "svkbd.pressbackground", "*", &type, &xval) && !colors[SchemePress][ColBg])
+			colors[SchemePress][ColBg] = estrdup(xval.addr);
+		if (XrmGetResource(xdb, "svkbd.pressforeground", "*", &type, &xval) && !colors[SchemePress][ColFg])
+			colors[SchemePress][ColFg] = estrdup(xval.addr);
+
+		if (XrmGetResource(xdb, "svkbd.pressshiftbackground", "*", &type, &xval) && !colors[SchemePressShift][ColBg])
+			colors[SchemePressShift][ColBg] = estrdup(xval.addr);
+		if (XrmGetResource(xdb, "svkbd.pressshiftforeground", "*", &type, &xval) && !colors[SchemePressShift][ColFg])
+			colors[SchemePressShift][ColFg] = estrdup(xval.addr);
+
+		if (XrmGetResource(xdb, "svkbd.highlightbackground", "*", &type, &xval) && !colors[SchemeHighlight][ColBg])
+			colors[SchemeHighlight][ColBg] = estrdup(xval.addr);
+		if (XrmGetResource(xdb, "svkbd.highlightforeground", "*", &type, &xval) && !colors[SchemeHighlight][ColFg])
+			colors[SchemeHighlight][ColFg] = estrdup(xval.addr);
+
+		if (XrmGetResource(xdb, "svkbd.highlightshiftbackground", "*", &type, &xval) && !colors[SchemeHighlightShift][ColBg])
+			colors[SchemeHighlightShift][ColBg] = estrdup(xval.addr);
+		if (XrmGetResource(xdb, "svkbd.highlightshiftforeground", "*", &type, &xval) && !colors[SchemeHighlightShift][ColFg])
+			colors[SchemeHighlightShift][ColFg] = estrdup(xval.addr);
+
+		if (XrmGetResource(xdb, "svkbd.overlaybackground", "*", &type, &xval) && !colors[SchemeOverlay][ColBg])
+			colors[SchemeOverlay][ColBg] = estrdup(xval.addr);
+		if (XrmGetResource(xdb, "svkbd.overlayforeground", "*", &type, &xval) && !colors[SchemeOverlay][ColFg])
+			colors[SchemeOverlay][ColFg] = estrdup(xval.addr);
+
+		if (XrmGetResource(xdb, "svkbd.overlayshiftbackground", "*", &type, &xval) && !colors[SchemeOverlayShift][ColBg])
+			colors[SchemeOverlayShift][ColBg] = estrdup(xval.addr);
+		if (XrmGetResource(xdb, "svkbd.overlayshiftforeground", "*", &type, &xval) && !colors[SchemeOverlayShift][ColFg])
+			colors[SchemeOverlayShift][ColFg] = estrdup(xval.addr);
+
+		if (XrmGetResource(xdb, "svkbd.windowbackground", "*", &type, &xval) && !colors[SchemeWindow][ColBg])
+			colors[SchemeWindow][ColBg] = estrdup(xval.addr);
+		if (XrmGetResource(xdb, "svkbd.windowforeground", "*", &type, &xval) && !colors[SchemeWindow][ColFg])
+			colors[SchemeWindow][ColFg] = estrdup(xval.addr);
+
+		XrmDestroyDatabase(xdb);
+	}
+}
+
 void
 setup(void)
 {
@@ -620,8 +827,23 @@ setup(void)
 		sh = DisplayHeight(dpy, screen);
 	}
 	drw = drw_create(dpy, screen, root, sw, sh);
-	if (!drw_fontset_create(drw, fonts, LENGTH(fonts)))
+
+	readxresources();
+
+	/* Apply defaults to font and colors*/
+	if (!fonts[0])
+		fonts[0] = estrdup(defaultfonts[0]);
+	for (i = 0; i < SchemeLast; ++i) {
+		for (j = 0; j < 2; ++j) {
+			if (!colors[i][j])
+				colors[i][j] = estrdup(defaultcolors[i][j]);
+		}
+	}
+
+	if (!drw_fontset_create(drw, (const char **) fonts, LENGTH(fonts)))
 		die("no fonts could be loaded");
+	free(fonts[0]);
+
 	drw_setscheme(drw, scheme[SchemeNorm]);
 
 	/* find an unused keycode to use as a temporary keycode (derived from source:
@@ -652,7 +874,12 @@ setup(void)
 
 	/* init appearance */
 	for (j = 0; j < SchemeLast; j++)
-		scheme[j] = drw_scm_create(drw, colors[j], 2);
+		scheme[j] = drw_scm_create(drw, (const char **) colors[j], 2);
+
+	for (j = 0; j < SchemeLast; ++j) {
+		free(colors[j][ColFg]);
+		free(colors[j][ColBg]);
+	}
 
 	/* init atoms */
 	if (isdock) {
@@ -705,8 +932,7 @@ setup(void)
 	ch->res_class = name;
 	ch->res_name = name;
 
-	XSetWMProperties(dpy, win, &str, &str, NULL, 0, sizeh, wmh,
-			ch);
+	XSetWMProperties(dpy, win, &str, &str, NULL, 0, sizeh, wmh, ch);
 
 	XFree(keysyms);
 	XFree(ch);
@@ -731,18 +957,21 @@ void
 updatekeys(void)
 {
 	int i, j;
-	int x = 0, y = 0, h, base, r = rows;
+	double base;
+	int x, y = 0, h, r = rows;
 
 	h = (wh - 1) / rows;
 	for (i = 0; i < numkeys; i++, r--) {
 		for (j = i, base = 0; j < numkeys && keys[j].keysym != 0; j++)
 			base += keys[j].width;
 		for (x = 0; i < numkeys && keys[i].keysym != 0; i++) {
-			keys[i].x = x;
-			keys[i].y = y;
-			keys[i].w = keys[i].width * (ww - 1) / base;
+			keys[i].x = x + xspacing;
+			keys[i].y = y + yspacing;
+			keys[i].w = keys[i].width * ww / base;
 			keys[i].h = r == 1 ? wh - y - 1 : h;
 			x += keys[i].w;
+			keys[i].w = keys[i].w - (xspacing * 2);
+			keys[i].h = keys[i].h - (yspacing * 2);
 		}
 		if (base != 0)
 			keys[i - 1].w = ww - 1 - keys[i - 1].x;
@@ -753,15 +982,19 @@ updatekeys(void)
 void
 usage(char *argv0)
 {
-	fprintf(stderr, "usage: %s [-hdvDO] [-g geometry] [-fn font] [-l layers] [-s initial_layer]\n", argv0);
+	fprintf(stderr, "usage: %s [-hdnovDOR] [-g geometry] [-fn font] [-l layers] [-s initial_layer]\n", argv0);
 	fprintf(stderr, "Options:\n");
 	fprintf(stderr, "  -d         - Set Dock Window Type\n");
 	fprintf(stderr, "  -D         - Enable debug\n");
 	fprintf(stderr, "  -O         - Disable overlays\n");
+	fprintf(stderr, "  -R         - Disable press-on-release\n");
+	fprintf(stderr, "  -n         - Do not simulate key presses for X\n");
+	fprintf(stderr, "  -o         - Print to standard output\n");
 	fprintf(stderr, "  -l         - Comma separated list of layers to enable\n");
 	fprintf(stderr, "  -s         - Layer to select on program start\n");
-	fprintf(stderr, "  -H [int]   - Height fraction, one key row takes 1/x of the screen height");
+	fprintf(stderr, "  -H [int]   - Height fraction, one key row takes 1/x of the screen height\n");
 	fprintf(stderr, "  -fn [font] - Set font (Xft, e.g: DejaVu Sans:bold:size=20)\n");
+	fprintf(stderr, "  -g         - Set the window position or size using the X geometry format\n");
 	exit(1);
 }
 
@@ -770,6 +1003,7 @@ setlayer(void)
 {
 	numkeys = countkeys(layers[currentlayer]);
 	memcpy(&keys, layers[currentlayer], sizeof(Key) * numkeys);
+	countrows();
 }
 
 void
@@ -778,7 +1012,7 @@ cyclelayer(void)
 	currentlayer++;
 	if (currentlayer >= numlayers)
 		currentlayer = 0;
-	if (debug) printdbg("Cycling to layer %d\n", currentlayer);
+	printdbg("Cycling to layer %d\n", currentlayer);
 	setlayer();
 	updatekeys();
 	drawkeyboard();
@@ -792,7 +1026,7 @@ togglelayer(void)
 	} else if (numlayers > 1) {
 		currentlayer = 1;
 	}
-	if (debug) printdbg("Toggling layer %d\n", currentlayer);
+	printdbg("Toggling layer %d\n", currentlayer);
 	setlayer();
 	updatekeys();
 	drawkeyboard();
@@ -801,25 +1035,40 @@ togglelayer(void)
 void
 showoverlay(int idx)
 {
-	if (debug) printdbg("Showing overlay %d\n", idx);
-	int i,j;
+	int i, j;
+
+	printdbg("Showing overlay %d\n", idx);
 
 	/* unpress existing key (visually only) */
 	for (i = 0; i < numkeys; i++) {
 		if (keys[i].pressed && !IsModifierKey(keys[i].keysym)) {
 			keys[i].pressed = 0;
-			drawkey(&keys[i]);
+			drawkey(&keys[i], True);
 			break;
 		}
 	}
 
-	for (i = idx, j=0; i < OVERLAYS; i++, j++) {
-		if (overlay[i].keysym == XK_Cancel)
+	for (i = idx, j = 0; i < OVERLAYS; i++, j++) {
+		if (overlay[i].keysym == XK_Cancel) {
 			break;
-		while (keys[j].keysym == 0)
+		}
+		/* certain modifier keys and basic keys are excluded from being overlayed: */
+		while (keys[j].keysym == 0 || keys[j].keysym == XK_Shift_L ||
+				keys[j].keysym == XK_Shift_R || keys[j].keysym == XK_Control_L ||
+				keys[j].keysym == XK_Control_R || keys[j].keysym == XK_Alt_L ||
+				keys[j].keysym == XK_Alt_R || keys[j].keysym == XK_BackSpace ||
+				keys[j].keysym == XK_Return || keys[j].keysym == XK_space ||
+				keys[j].keysym == XK_Cancel)
 			j++;
+		if (overlay[i].width > 1)
+			j += overlay[i].width - 1;
+		if (j >= numkeys)
+			break;
 		keys[j].label = overlay[i].label;
+		keys[j].label2 = overlay[i].label2;
 		keys[j].keysym = overlay[i].keysym;
+		keys[j].modifier = overlay[i].modifier;
+		keys[j].isoverlay = True;
 	}
 	currentoverlay = idx;
 	overlaykeysym = ispressingkeysym;
@@ -832,10 +1081,10 @@ showoverlay(int idx)
 void
 hideoverlay(void)
 {
-	if (debug) printdbg("Hiding overlay, overlay was #%d\n", currentoverlay);
+	printdbg("Hiding overlay, overlay was #%d\n", currentoverlay);
 	currentoverlay = -1;
 	overlaykeysym = 0;
-	currentlayer = -1;
+	currentlayer--;
 	cyclelayer();
 }
 
@@ -844,7 +1093,7 @@ sigterm(int signo)
 {
 	running = False;
 	sigtermd = True;
-	if (debug) printdbg("SIGTERM received\n");
+	printdbg("SIGTERM received\n");
 }
 
 void
@@ -867,7 +1116,7 @@ init_layers(char *layer_names_list, const char *initial_layer_name)
 	} else {
 		s = strtok(layer_names_list, ",");
 		while (s != NULL) {
-			if (numlayers+1 > LAYERS)
+			if (numlayers + 1 > LAYERS)
 				die("too many layers specified");
 			found = 0;
 			for (j = 0; j < LAYERS; j++) {
@@ -895,6 +1144,9 @@ init_layers(char *layer_names_list, const char *initial_layer_name)
 void
 printdbg(const char *fmt, ...)
 {
+	if (!debug)
+		return;
+
 	va_list ap;
 	va_start(ap, fmt);
 	vfprintf(stderr, fmt, ap);
@@ -919,26 +1171,26 @@ main(int argc, char *argv[])
 		if ((tmp = getenv("SVKBD_ENABLEOVERLAYS")))
 			enableoverlays = atoi(tmp);
 	}
-	if ((tmp = getenv("SVKBD_LAYERS"))) {
-		if (!(layer_names_list = strdup(tmp)))
-			die("memory allocation error");
-	}
+	if ((tmp = getenv("SVKBD_LAYERS")))
+		layer_names_list = estrdup(tmp);
 
 	if ((tmp = getenv("SVKBD_HEIGHTFACTOR")))
 		heightfactor = atoi(tmp);
 
+	if ((tmp = getenv("SVKBD_PRESSONRELEASE"))) /* defaults to 1 */
+		pressonrelease = atoi(tmp);
+
 	/* parse command line arguments */
 	for (i = 1; argv[i]; i++) {
 		if (!strcmp(argv[i], "-v")) {
 			die("svkbd-"VERSION);
 		} else if (!strcmp(argv[i], "-d")) {
 			isdock = True;
-			continue;
 		} else if (!strncmp(argv[i], "-g", 2)) {
 			if (i >= argc - 1)
-				continue;
+				usage(argv[0]);
 
-			bitm = XParseGeometry(argv[i + 1], &xr, &yr, &wr, &hr);
+			bitm = XParseGeometry(argv[++i], &xr, &yr, &wr, &hr);
 			if (bitm & XValue)
 				wx = xr;
 			if (bitm & YValue)
@@ -951,35 +1203,44 @@ main(int argc, char *argv[])
 				wx = -1;
 			if (bitm & YNegative && wy == 0)
 				wy = -1;
-			i++;
 		} else if (!strcmp(argv[i], "-fn")) { /* font or font set */
-			fonts[0] = argv[++i];
+			if (i >= argc - 1)
+				usage(argv[0]);
+			fonts[0] = estrdup(argv[++i]);
 		} else if (!strcmp(argv[i], "-D")) {
 			debug = 1;
 		} else if (!strcmp(argv[i], "-h")) {
 			usage(argv[0]);
 		} else if (!strcmp(argv[i], "-O")) {
 			enableoverlays = 0;
+		} else if (!strcmp(argv[i], "-o")) {
+			printoutput = 1;
+		} else if (!strcmp(argv[i], "-n")) {
+			simulateoutput = 0;
+		} else if (!strcmp(argv[i], "-R")) {
+			pressonrelease = 0;
 		} else if (!strcmp(argv[i], "-l")) {
 			if (i >= argc - 1)
-				continue;
+				usage(argv[0]);
 			free(layer_names_list);
-			if (!(layer_names_list = strdup(argv[++i])))
-				die("memory allocation error");
+			layer_names_list = estrdup(argv[++i]);
 		} else if (!strcmp(argv[i], "-s")) {
 			if (i >= argc - 1)
-				continue;
+				usage(argv[0]);
 			initial_layer_name = argv[++i];
 		} else if (!strcmp(argv[i], "-H")) {
 			if (i >= argc - 1)
-				continue;
+				usage(argv[0]);
 			heightfactor = atoi(argv[++i]);
 		} else {
 			fprintf(stderr, "Invalid argument: %s\n", argv[i]);
-			exit(2);
+			usage(argv[0]);
 		}
 	}
 
+	if (printoutput)
+		setbuf(stdout, NULL); /* unbuffered output */
+
 	if (heightfactor <= 0)
 		die("height factor must be a positive integer");
 
